// var WebFont = require("webfontloader");
var _ = require("lodash");
var _fvd = require("fvd");
var format = require("string-format");

/**
 * The threshold to determine non-white pixels.
 *
 * @type Number
 */
var threshold = 0.75;

/**
 * The device pixel ratio, used to calculate the dimensions on high-DPI screens.
 *
 * @type Number
 */
var dpr = window.devicePixelRatio;

/**
 * The element that will contain the canvases.
 *
 * @type Node
 */
var container;

module.exports = {
  calculate: calculate,

  /**
   * Event listener for the fontactive event of the Web Font Loader.
   *
   * Calculate the cap-height and execute the given callback with the resulted properties.
   *
   * @param  Function callback
   * @param  String   text
   * @return Function
   */
  fontActive: function(callback, text) {
    return function(familyName, fvd) {
      // Parse the fvd properties returned from the Web Font Loader
      var parsed = _fvd.parse(fvd);

      // Add the font-family
      _.set(parsed, "font-family", familyName);

      // Calculate the cap-height
      var properties = calculate(parsed, text);

      // Execute the callback
      return callback(properties);
    };
  },

  /**
   * Set a container element to display the rendered canvas.
   *
   * @param  Node element
   * @return Node
   */
  setContainer: function(element) {
    return container = element;
  }
};

/**
 * Calculate the cap height.
 *
 * @param Object properties
 * @param String text
 * @return Object
 */
function calculate(properties, text) {
  // Merge with the default properties
  properties = mergeWithDefaults(properties);

  // Measure the actual size of the text
  var height = measureTextHeight(properties, text);

  // Calculate the ratio between the height and the font-size
  var capHeight = height / getFontSize(properties);

  // Add the cap-height to the properties
  _.set(properties, "cap-height", capHeight);

  return properties;
}

/**
 * Merge the given properties with the defaults.
 *
 * @param  Object properties
 * @return Object
 */
function mergeWithDefaults(properties) {
  var defaults = {
    "font-style": "normal",
    "font-weight": 400,
    "font-size": "100px",
    "font-family": "serif"
  };

  return _.merge(defaults, properties);
}

/**
 * Measure the height of the given text with the specified font properties.
 *
 * @param  Object properties
 * @param  String text
 * @return Number
 *
 * @throws RangeError
 */
function measureTextHeight(properties, text) {
  var fontSize = getFontSize(properties);

  text = text || "H";

  if (/\s/.test(text)) {
    throw new RangeError("Cannot calculate the height of whitespace.");
  }

  var base = 2 * fontSize;

  // Set the actual pixel dimensions
  var dimensions = {
    width: base * text.length,
    height: base
  };

  // Set the physical pixel dimensions
  var physicalDimensions = _.mapValues(dimensions, toDot);

  var canvas = createCanvas(dimensions, physicalDimensions);
  var context = getContext(canvas);

  context = drawContext(context, properties, dimensions, text);

  var pixelData = getPixelData(context, physicalDimensions);
  var ascent = getAscent(pixelData, physicalDimensions.width);
  var descent = getDescent(pixelData, physicalDimensions.width);
  var capHeight = getHeight(descent, ascent);

  // Draw a rectangle marking the cap height bounds
  context.fillStyle = "rgba(255, 85, 51, .1)";
  context.fillRect(0, toPx(ascent), toPx(physicalDimensions.width), toPx(capHeight));

  // Display the canvas
  displayCanvas(canvas);

  return toPx(capHeight);
}

/**
 * Get the font-size as a number.
 *
 * @param Object properties
 * @return Number
 */
function getFontSize(properties) {
  var fontSize = parseFloat(_.replace(_.get(properties, "font-size", 0), /\D/g, ''));

  if (isNaN(fontSize)) {
    throw new RangeError("Cannot calculate the height of the text with invalid font-size.");
  }

  if (fontSize === 0) {
    throw new RangeError("Cannot calculate the height of the text with font-size: 0.");
  }

  return fontSize;
}

/**
 * Create a canvas with the given dimensions.
 *
 * @param Object dimensions
 * @param Object physicalDimensions
 * @return HTMLCanvasElement
 */
function createCanvas(dimensions, physicalDimensions) {
  var canvas = document.createElement("canvas");

  canvas.width = physicalDimensions.width;
  canvas.height = physicalDimensions.height;

  canvas.style.width = dimensions.width + "px";
  canvas.style.height = dimensions.height + "px";

  canvas.style.margin = "10px";

  return canvas;
}

/**
 * Append the given canvas to the container element.
 *
 * @param HTMLCanvasElement canvas
 * @return void
 */
function displayCanvas(canvas) {
  if (container && container.nodeType) {
    container.appendChild(canvas);
  }
}

/**
 * Get the canvas drawing context.
 *
 * @return CanvasRenderingContext2D
 */
function getContext(canvas) {
  var context = canvas.getContext("2d");

  context.scale(dpr, dpr);

  return context;
}

/**
 * Draw the text and the background in the canvas context.
 *
 * @param CanvasRenderingContext2D context
 * @param Object properties
 * @param Object dimensions
 * @param String context
 * @return CanvasRenderingContext2D
 */
function drawContext(context, properties, dimensions, text) {
  context.font = composeFontProperty(properties);

  // Align the text horizontally and vertically
  context.textAlign = "center";
  context.textBaseline = "middle";

  // Set all canvas pixeldata values to 255, with all the content
  // data being 0. This lets us scan for data[i] != 255.
  context.fillStyle = "white";
  context.fillRect(0, 0, dimensions.width, dimensions.height);
  context.fillStyle = "black";

  // Draw the text at the center
  context.fillText(text, dimensions.width / 2, dimensions.height / 2);

  return context;
}

/**
 * Compose the font shorthand property.
 *
 * @param Object properties
 * @return String
 *
 * @see https://developer.mozilla.org/en-US/docs/Web/CSS/font
 */
function composeFontProperty(properties) {
  return format("{font-style} {font-weight} {font-size} {font-family}", properties);
}

/**
 * Get the pixel data of the context.
 *
 * @param  CanvasRenderingContext2D context
 * @param  Object dimensions
 * @return Uint8ClampedArray
 */
function getPixelData(context, dimensions) {
  return context.getImageData(0, 0, dimensions.width, dimensions.height).data;
}

/**
 * Calculate the height of the text with the given ascent and descent.
 *
 * @param  Number ascent
 * @param  Number descent
 * @return Number
 */
function getHeight(descent, ascent) {
  return descent + 1 - ascent;
}

/**
 * Get the ascent of the text.
 *
 * Forward scan the pixel data and get the vertical position of the first non-white pixel.
 *
 * @param  Uint8ClampedArray pixelData
 * @param  Number width
 * @return Number
 */
function getAscent(pixelData, width) {
  var i = _.findIndex(pixelData, isNotWhite);

  return getYPosition(i, width);
}

/**
 * Get the descent of the text.
 *
 * Reverse scan the pixel data and get the vertical position of the last non-white pixel.
 *
 * @param  Uint8ClampedArray pixelData
 * @param  Number width
 * @return Number
 */
function getDescent(pixelData, width) {
  var i = _.findLastIndex(pixelData, isNotWhite);

  return getYPosition(i, width);
}

/**
 * Get the vertical position of the pixel data index, given the width.
 *
 * @param Number i
 * @param Number width
 * @return Number
 */
function getYPosition(i, width) {
  return _.floor((i / 4) / width);
}

/**
 * Check the given RGBA value is below the threshold.
 *
 * @param Number value
 * @return Boolean
 */
function isNotWhite(value) {
  return value < (255 * threshold);
}

/**
 * Convert a physical pixel value to a CSS pixel
 */
function toPx(value) {
  return value / dpr;
}

/**
 * Convert a CSS pixel to a physical pixel
 */
function toDot(value) {
  return value * dpr;
}
